﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.ComponentModel;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.OptimizationExpertSystem.Views {
  [View("Expert-system Optimizer View")]
  [Content(typeof(ExpertSystem), IsDefaultView = true)]
  public partial class ExpertSystemView : NamedItemView {
    protected TypeSelectorDialog problemTypeSelectorDialog;
    protected virtual bool SuppressEvents { get; set; }
    private bool okbDownloadInProgress;

    public new ExpertSystem Content {
      get { return (ExpertSystem)base.Content; }
      set { base.Content = value; }
    }

    public ExpertSystemView() {
      InitializeComponent();
      refreshMapButton.Text = string.Empty;
      refreshMapButton.Image = VSImageLibrary.Refresh;
    }

    protected override void Dispose(bool disposing) {
      if (disposing) {
        if (problemTypeSelectorDialog != null) problemTypeSelectorDialog.Dispose();
        if (components != null) components.Dispose();
      }
      base.Dispose(disposing);
    }

    protected override void DeregisterContentEvents() {
      Content.PropertyChanged -= Content_PropertyChanged;
      Content.SuggestedInstances.CollectionReset -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsAdded -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsMoved -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsRemoved -= SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsReplaced -= SuggestedInstancesOnChanged;
      base.DeregisterContentEvents();
    }
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.PropertyChanged += Content_PropertyChanged;
      Content.SuggestedInstances.CollectionReset += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsAdded += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsMoved += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsRemoved += SuggestedInstancesOnChanged;
      Content.SuggestedInstances.ItemsReplaced += SuggestedInstancesOnChanged;
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      SuppressEvents = true;
      okbDownloadInProgress = false;
      try {
        if (Content == null) {
          maxEvaluationsTextBox.Text = String.Empty;

          problemViewHost.Content = null;
          algorithmViewHost.Content = null;
          runsView.Content = null;
          algorithmInstancesViewHost.Content = null;
          problemInstancesView.Content = null;
        } else {
          maxEvaluationsTextBox.Text = Content.MaximumEvaluations.ToString();
          problemViewHost.Content = Content.Problem;
          runsView.Content = Content.Runs;
          algorithmInstancesViewHost.Content = Content.AlgorithmInstances;
          problemInstancesView.Content = Content.ProblemInstances;
        }
      } finally { SuppressEvents = false; }
      UpdateSuggestedInstancesCombobox();
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      maxEvaluationsTextBox.Enabled = Content != null && !ReadOnly && !Locked;
      newProblemButton.Enabled = Content != null && !ReadOnly && !Locked;
      openProblemButton.Enabled = Content != null && !ReadOnly && !Locked;
      problemViewHost.Enabled = Content != null && !ReadOnly && !Locked;
      suggestedInstancesComboBox.Enabled = Content != null && !ReadOnly && !Locked;
      algorithmViewHost.Enabled = Content != null && !ReadOnly && !Locked;
      runsView.Enabled = Content != null;
      algorithmInstancesViewHost.Enabled = Content != null && !okbDownloadInProgress;
      refreshMapButton.Enabled = Content != null;
      okbDownloadButton.Enabled = Content != null && Content.Problem != null && !ReadOnly && !Locked && !okbDownloadInProgress;
    }

    private void UpdateSuggestedInstancesCombobox() {
      var prevSelection = (IAlgorithm)suggestedInstancesComboBox.SelectedItem;
      var prevNewIndex = -1;
      suggestedInstancesComboBox.Items.Clear();
      if (Content == null) return;

      for (var i = 0; i < Content.SuggestedInstances.Count; i++) {
        suggestedInstancesComboBox.Items.Add(Content.SuggestedInstances[i]);
        if (prevSelection == null || Content.SuggestedInstances[i].Name == prevSelection.Name)
          prevNewIndex = prevSelection == null ? 0 : i;
      }
      if (prevNewIndex >= 0) {
        suggestedInstancesComboBox.SelectedIndex = prevNewIndex;
      }
    }

    #region Event Handlers
    #region Content events
    private void Content_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      if (InvokeRequired) {
        Invoke((Action<object, PropertyChangedEventArgs>)Content_PropertyChanged, sender, e);
        return;
      }
      SuppressEvents = true;
      try {
        switch (e.PropertyName) {
          case "AlgorithmInstances": algorithmInstancesViewHost.Content = Content.AlgorithmInstances; break;
          case "MaximumEvaluations": maxEvaluationsTextBox.Text = Content.MaximumEvaluations.ToString(); break;
          case "Problem": problemViewHost.Content = Content.Problem; break;
          case "ProblemInstances": problemInstancesView.Content = Content.ProblemInstances; break;
        }
      } finally { SuppressEvents = false; }
      SetEnabledStateOfControls();
    }

    private void SuggestedInstancesOnChanged(object sender, EventArgs e) {
      UpdateSuggestedInstancesCombobox();
    }
    #endregion

    #region Control events
    private void maxEvaluationsTextBox_Validating(object sender, CancelEventArgs e) {
      if (SuppressEvents) return;
      if (InvokeRequired) {
        Invoke((Action<object, CancelEventArgs>)maxEvaluationsTextBox_Validating, sender, e);
        return;
      }
      int value;
      if (!int.TryParse(maxEvaluationsTextBox.Text, out value)) {
        e.Cancel = !maxEvaluationsTextBox.ReadOnly && maxEvaluationsTextBox.Enabled;
        errorProvider.SetError(maxEvaluationsTextBox, "Please enter a valid integer number.");
      } else {
        Content.MaximumEvaluations = value;
        e.Cancel = false;
        errorProvider.SetError(maxEvaluationsTextBox, null);
      }
    }

    private void newProblemButton_Click(object sender, EventArgs e) {
      if (problemTypeSelectorDialog == null) {
        problemTypeSelectorDialog = new TypeSelectorDialog { Caption = "Select Problem" };
        problemTypeSelectorDialog.TypeSelector.Caption = "Available Problems";
        problemTypeSelectorDialog.TypeSelector.Configure(typeof(ISingleObjectiveHeuristicOptimizationProblem)
          , showNotInstantiableTypes: false, showGenericTypes: false);
      }
      if (problemTypeSelectorDialog.ShowDialog(this) == DialogResult.OK) {
        try {
          Content.Problem = (ISingleObjectiveHeuristicOptimizationProblem)problemTypeSelectorDialog.TypeSelector.CreateInstanceOfSelectedType();
        } catch (Exception ex) {
          ErrorHandling.ShowErrorDialog(this, ex);
        }
      }
    }

    private void openProblemButton_Click(object sender, EventArgs e) {
      openFileDialog.Title = "Open Problem";
      if (openFileDialog.ShowDialog(this) == DialogResult.OK) {
        newProblemButton.Enabled = openProblemButton.Enabled = false;
        problemViewHost.Enabled = false;

        ContentManager.LoadAsync(openFileDialog.FileName, delegate(IStorableContent content, Exception error) {
          try {
            if (error != null) throw error;
            var problem = content as ISingleObjectiveHeuristicOptimizationProblem;
            if (problem == null) {
              var algorithm = content as IAlgorithm;
              if (algorithm == null || !(algorithm.Problem is ISingleObjectiveHeuristicOptimizationProblem))
                MessageBox.Show(this, "The selected file is not a problem, nor an algorithm with a problem.", "Invalid File", MessageBoxButtons.OK, MessageBoxIcon.Error);
              else Content.Problem = (ISingleObjectiveHeuristicOptimizationProblem)algorithm.Problem;
            } else
              Content.Problem = problem;
          } catch (Exception ex) {
            ErrorHandling.ShowErrorDialog(this, ex);
          } finally {
            Invoke(new Action(delegate() {
              problemViewHost.Enabled = true;
              newProblemButton.Enabled = openProblemButton.Enabled = true;
            }));
          }
        });
      }
    }

    private void problemTabPage_DragEnterOver(object sender, DragEventArgs e) {
      e.Effect = DragDropEffects.None;
      var prob = e.Data.GetData(Constants.DragDropDataFormat) as ISingleObjectiveHeuristicOptimizationProblem;
      if (!ReadOnly && prob != null) {
        if ((e.KeyState & 32) == 32) e.Effect = DragDropEffects.Link;  // ALT key
        else if ((e.KeyState & 4) == 4) e.Effect = DragDropEffects.Move;  // SHIFT key
        else if (e.AllowedEffect.HasFlag(DragDropEffects.Copy)) e.Effect = DragDropEffects.Copy;
        else if (e.AllowedEffect.HasFlag(DragDropEffects.Move)) e.Effect = DragDropEffects.Move;
        else if (e.AllowedEffect.HasFlag(DragDropEffects.Link)) e.Effect = DragDropEffects.Link;
      }
    }

    private void problemTabPage_DragDrop(object sender, DragEventArgs e) {
      if (e.Effect != DragDropEffects.None) {
        var prob = e.Data.GetData(Constants.DragDropDataFormat) as ISingleObjectiveHeuristicOptimizationProblem;
        if (e.Effect.HasFlag(DragDropEffects.Copy)) prob = (ISingleObjectiveHeuristicOptimizationProblem)prob.Clone();
        Content.Problem = prob;
      }
    }

    private void suggestedInstancesComboBox_SelectedIndexChanged(object sender, EventArgs e) {
      if (SuppressEvents) return;
      if (InvokeRequired) { Invoke((Action<object, EventArgs>)suggestedInstancesComboBox_SelectedIndexChanged, sender, e); return; }
      if (suggestedInstancesComboBox.SelectedIndex >= 0)
        algorithmViewHost.Content = (IAlgorithm)suggestedInstancesComboBox.SelectedItem;
      else algorithmViewHost.Content = null;
    }
    #endregion

    private bool validDragOperation;
    private void instancesDropPanel_DragEnter(object sender, DragEventArgs e) {
      validDragOperation = false;
      if (!ReadOnly && (e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat) is IRun)) {
        validDragOperation = true;
      } else if (!ReadOnly && (e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat) is IEnumerable)) {
        validDragOperation = true;
        var items = (IEnumerable)e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat);
        foreach (object item in items)
          validDragOperation = validDragOperation && (item is IRun);
      }
    }
    private void instancesDropPanel_DragOver(object sender, DragEventArgs e) {
      e.Effect = DragDropEffects.None;
      if (!validDragOperation) return;
      if ((e.KeyState & 32) == 32) e.Effect = DragDropEffects.Link;  // ALT key
      else if ((e.KeyState & 4) == 4) e.Effect = DragDropEffects.Move;  // SHIFT key
      else if (e.AllowedEffect.HasFlag(DragDropEffects.Copy)) e.Effect = DragDropEffects.Copy;
      else if (e.AllowedEffect.HasFlag(DragDropEffects.Move)) e.Effect = DragDropEffects.Move;
      else if (e.AllowedEffect.HasFlag(DragDropEffects.Link)) e.Effect = DragDropEffects.Link;
    }
    private void instancesDropPanel_DragDrop(object sender, DragEventArgs e) {
      if (e.Effect == DragDropEffects.None) return;
      if (e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat) is IRun) {
        var item = (IRun)e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat);
        Content.ProblemInstances.Add(e.Effect.HasFlag(DragDropEffects.Copy) ? (IRun)item.Clone() : item);
      } else if (e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat) is IEnumerable) {
        var items = ((IEnumerable)e.Data.GetData(HeuristicLab.Common.Constants.DragDropDataFormat)).Cast<IRun>();
        if (e.Effect.HasFlag(DragDropEffects.Copy)) {
          var cloner = new Cloner();
          items = items.Select(x => cloner.Clone(x));
        }
        Content.ProblemInstances.AddRange(items);
      }
    }
    #endregion

    private void refreshMapButton_Click(object sender, EventArgs e) {
      Content.UpdateInstanceMap();
    }

    private void okbDownloadButton_Click(object sender, EventArgs e) {
      var progress = new Progress();
      progress.ProgressStateChanged += OkbDownloadProgressOnStateChanged;
      Content.UpdateKnowledgeBaseAsync(progress);
      MainFormManager.GetMainForm<MainForm.WindowsForms.MainForm>().AddOperationProgressToView(progressPanel, progress);
      progressPanel.Visible = true;
      SetEnabledStateOfControls();
    }

    private void OkbDownloadProgressOnStateChanged(object sender, EventArgs eventArgs) {
      var progress = (IProgress)sender;
      okbDownloadInProgress = progress.ProgressState == ProgressState.Started;
      SetEnabledStateOfControls();
      if (!okbDownloadInProgress) {
        progressPanel.Visible = false;
        progress.ProgressStateChanged -= OkbDownloadProgressOnStateChanged;
      }
    }
  }
}
